"""WS-694-state-machine-schema

Revision ID: 531f4196d0ec
Revises: 4e009168b3bf
Create Date: 2021-09-24 14:42:31.091988

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "531f4196d0ec"
down_revision = "4e009168b3bf"
branch_labels = None
depends_on = None


"""
This migration is morally equivalent to this set of SQL:

    CREATE TABLE capability_states (
        capability_name VARCHAR REFERENCES capabilities ON UPDATE CASCADE ON DELETE CASCADE, 
        state VARCHAR, 
        PRIMARY KEY (capability_name, state)
    );
    
    CREATE TABLE capability_state_transitions (
        transition_id SERIAL PRIMARY KEY,
        capability_name VARCHAR,
        from_state VARCHAR,
        pattern JSONPATH,
        to_state VARCHAR,
        CONSTRAINT from_state_fk FOREIGN KEY (capability_name, from_state) REFERENCES capability_states (capability_name, state) ON UPDATE CASCADE ON DELETE CASCADE,
        CONSTRAINT to_state_fk FOREIGN KEY (capability_name, to_state) REFERENCES capability_states (capability_name, state) ON UPDATE CASCADE ON DELETE CASCADE,
    );
    
    CREATE TABLE capability_state_actions (
        action_id SERIAL PRIMARY KEY,
        transition_id INTEGER NOT NULL REFERENCES capability_state_transitions ON UPDATE CASCADE ON DELETE CASCADE,
        action_type VARCHAR NOT NULL,
        arguments VARCHAR        
    );
"""

def create_capability_states():
    op.create_table(
        "capability_states",
        sa.Column(
            "capability_name",
            sa.String,
            sa.ForeignKey("capabilities.capability_name"),
            comment="The capability this state belongs to",
        ),
        sa.Column(
            "state",
            sa.String,
            nullable=False,
            comment="A state in this state machine, such as 'Executing' or 'Awaiting QA'",
        ),
        sa.PrimaryKeyConstraint("capability_name", "state"),
        comment="The states that capabilities can be in (per-capability states)",
    )


def create_capability_state_transitions():
    op.create_table(
        "capability_state_transitions",
        sa.Column(
            "transition_id",
            sa.Integer,
            primary_key=True,
            autoincrement=True,
            comment="The unique identifier of this state machine transition",
        ),
        sa.Column(
            "capability_name",
            sa.String,
            sa.ForeignKey("capabilities.capability_name"),
            nullable=False,
            comment="The capability to whose state machine this transition belongs",
        ),
        sa.Column(
            "from_state",
            sa.String,
            nullable=False,
            comment="The starting state for this transition",
        ),
        sa.Column(
            "pattern",
            sa.String,
            nullable=False,
            comment="A JSONPATH pattern for what events cause us to take this transition",
        ),
        sa.Column(
            "to_state",
            sa.String,
            nullable=False,
            comment="The destination state for this transition",
        ),
        sa.ForeignKeyConstraint(
            ("capability_name", "from_state"),
            ["capability_states.capability_name", "capability_states.state"],
            ondelete="CASCADE",
            onupdate="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ("capability_name", "to_state"),
            ["capability_states.capability_name", "capability_states.state"],
            ondelete="CASCADE",
            onupdate="CASCADE",
        ),
        comment="The transition tables for capability state machines",
    )


def create_capability_state_actions():
    op.create_table(
        "capability_state_actions",
        sa.Column(
            "action_id",
            sa.Integer,
            primary_key=True,
            autoincrement=True,
            comment="The unique identifier for this state action",
        ),
        sa.Column(
            "transition_id",
            sa.Integer,
            sa.ForeignKey("capability_state_transitions.transition_id"),
            nullable=False,
            comment="The transition this action is executed during",
        ),
        sa.Column(
            "action_type",
            sa.String,
            nullable=False,
            comment="The type of action this is. Most likely the name of an Action subclass such as SendNotification or ExecuteWorkflow",
        ),
        sa.Column(
            "arguments", sa.String, comment="Additional arguments for the action"
        ),
        comment="Actions that are taken during transitioning from one state to another in a capability state machine",
    )


def copy_step_sequences_to_states():
    """
    So the general procedure here is this sort of loop:

    For each step sequence that looks like "prepare-and-run-workflow foo; await-workflow",
      Create the states "Start", "Executing", "Complete" and "Failed"
      Create the transitions:
        - start -> executing on {started}
        - executing -> complete on {workflow-complete}
        - executing -> failed on {workflow-failed}
      Create the actions:
        - QueueWorkflow(foo) on start->executing
        - Notify("capability-complete") on executing->complete
        - Notify("capability-failed") on executing->failed

    Doing this in general is a real treat!
    :return:
    """
    sql = """
    INSERT INTO capability_states (capability_name, state) 
    SELECT capability_name, 'Start' as state FROM capabilities 
    WHERE capability_steps ~ E'^prepare-and-run-workflow (.*)\nawait-workflow$';
    INSERT INTO capability_states (capability_name, state) 
    SELECT capability_name, 'Executing' as state FROM capabilities 
    WHERE capability_steps ~ E'^prepare-and-run-workflow (.*)\nawait-workflow$';
    INSERT INTO capability_states (capability_name, state) 
    SELECT capability_name, 'Complete' as state FROM capabilities 
    WHERE capability_steps ~ E'^prepare-and-run-workflow (.*)\nawait-workflow$';
    INSERT INTO capability_states (capability_name, state) 
    SELECT capability_name, 'Failed' as state FROM capabilities 
    WHERE capability_steps ~ E'^prepare-and-run-workflow (.*)\nawait-workflow$';
    
    INSERT INTO capability_state_transitions (capability_name, from_state, pattern, to_state)
    SELECT capability_name, 'Start', 'exists($ ? (@.event_type == "start"))', 'Executing' FROM capabilities
    WHERE capability_steps ~ E'^prepare-and-run-workflow (.*)\nawait-workflow$';
    INSERT INTO capability_state_transitions (capability_name, from_state, pattern, to_state)
    SELECT capability_name, 'Executing', 'exists($ ? (@.event_type == "workflow-complete"))', 'Complete' FROM capabilities
    WHERE capability_steps ~ E'^prepare-and-run-workflow (.*)\nawait-workflow$';
    INSERT INTO capability_state_transitions (capability_name, from_state, pattern, to_state)
    SELECT capability_name, 'Executing', 'exists($ ? (@.event_type == "workflow-failed"))', 'Failed' FROM capabilities
    WHERE capability_steps ~ E'^prepare-and-run-workflow (.*)\nawait-workflow$';

    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'QueueWorkflow', (regexp_match(capability_steps, E'^prepare-and-run-workflow (.*)\nawait-workflow$'))[1] as arguments
    FROM capability_state_transitions cst
    JOIN capabilities c on cst.capability_name = c.capability_name
    WHERE cst.from_state = 'Start' and cst.to_state = 'Executing';
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendNotification', 'workflow-complete'
    FROM capability_state_transitions cst
    WHERE cst.from_state = 'Executing' and cst.to_state = 'Complete';    
    INSERT INTO capability_state_actions (transition_id, action_type, arguments)
    SELECT transition_id, 'SendNotification', 'workflow-failed'
    FROM capability_state_transitions cst
    WHERE cst.from_state = 'Executing' and cst.to_state = 'Failed';    
    """


def upgrade():
    """

    :return:
    """
    create_capability_states()
    create_capability_state_transitions()
    create_capability_state_actions()
    copy_step_sequences_to_states()


def downgrade():
    op.drop_table("capability_state_actions")
    op.drop_table("capability_state_transitions")
    op.drop_table("capability_states")
